/*
 * Kneron Example Post-Processing driver
 *
 * Copyright (C) 2018-2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "base.h"
#include "kdpio.h"
#include "ipc.h"
#include "dbg_print.h"
#include "kdrv_gdma.h"

#define YOLO_V3_O1_GRID_W 7
#define YOLO_V3_O1_GRID_H 7
#define YOLO_V3_O1_GRID_MAX (YOLO_V3_O1_GRID_W * YOLO_V3_O1_GRID_H)
#define YOLO_V3_O2_GRID_W 14
#define YOLO_V3_O2_GRID_H 14
#define YOLO_V3_O2_GRID_MAX (YOLO_V3_O2_GRID_W * YOLO_V3_O2_GRID_H)
#define YOLO_V3_CELL_BOX_NUM 3

#define YOLO_CELL_BOX_NUM 5
#define YOLO_CLASS_MAX 80
#define YOLO_GOOD_BOX_MAX 30
#define YOLO_BOX_FIX_CH 5 /* x, y, w, h, confidence score */
#define YOLO_MAX_DETECTION_PER_CLASS 20
#define YOLO_PROB_THRESH_POS 0

#define KDP_COL_MIN 16 /* Bytes, i.e. 128 bits */

static float prob_thresh_yolov3 = 0.2; // probability threshold for yolo v3
const float nms_thresh_yolov3_ud = 0.45;  // non max suppression threshold for yolo v3

const float anchers_v3[2][3][2] = {
    {{81, 82}, {135, 169}, {344, 319}},
    {{23, 27}, {37, 58}, {81, 82}}};

const float anchers_v5[3][3][2] = {
    {{10, 13}, {16, 30}, {33, 23}},
    {{30, 61}, {62, 45}, {59, 119}},
    {{116, 90}, {156, 198}, {373, 326}}};

/* Shared global variable area among models */
struct yolo_v3_post_globals_s
{
    float box_class_probs[YOLO_CLASS_MAX];
    struct bounding_box_s bboxes_v3[YOLO_V3_O2_GRID_MAX * YOLO_V3_CELL_BOX_NUM];
    struct bounding_box_s result_tmp_s[YOLO_GOOD_BOX_MAX];
};

/* Model globals */
static struct yolo_v3_post_globals_s *yolov3_gp;

static void *get_gp(void **gp, size_t len)
{
    if (*gp == NULL)
    {
        *gp = malloc(len);
        if (*gp == NULL)
            kerr_msg("ncpu: ERR: malloc %d in get_gp()\n", len);
    }

    return *gp;
}

static inline struct yolo_v3_post_globals_s *get_yolov3_gp(void)
{
    return (struct yolo_v3_post_globals_s *)get_gp((void **)&yolov3_gp, sizeof(struct yolo_v3_post_globals_s));
}


static float do_div_scale_optim(float v, float scale)
{
    return (v * scale);
}

static uint32_t round_up(uint32_t num)
{
    return ((num + (KDP_COL_MIN - 1)) & ~(KDP_COL_MIN - 1));
}

static float sigmoid(float x)
{
    float exp_value;
    float return_value;

    exp_value = expf(-x);

    return_value = 1 / (1 + exp_value);

    return return_value;
}

static int float_comparator(float a, float b)
{
    float diff = a - b;

    if (diff < 0)
        return 1;
    else if (diff > 0)
        return -1;
    return 0;
}

static int box_score_comparator(const void *pa, const void *pb)
{
    float a, b;

    a = ((struct bounding_box_s *)pa)->score;
    b = ((struct bounding_box_s *)pb)->score;

    return float_comparator(a, b);
}

static float overlap(float l1, float r1, float l2, float r2)
{
    float left = l1 > l2 ? l1 : l2;
    float right = r1 < r2 ? r1 : r2;
    return right - left;
}

static float box_intersection(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float w, h, area;

    w = overlap(a->x1, a->x2, b->x1, b->x2);
    h = overlap(a->y1, a->y2, b->y1, b->y2);

    if (w < 0 || h < 0)
        return 0;

    area = w * h;
    return area;
}

static float box_union(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float i, u;

    i = box_intersection(a, b);
    u = (a->y2 - a->y1) * (a->x2 - a->x1) + (b->y2 - b->y1) * (b->x2 - b->x1) - i;

    return u;
}

static float box_iou(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float c;

    if (box_intersection(a, b) / box_intersection(a, a) > box_intersection(a, b) / box_intersection(b, b))
        c = box_intersection(a, b) / box_intersection(a, a);
    else
        c = box_intersection(a, b) / box_intersection(b, b);

    if (c < box_intersection(a, b) / box_union(a, b))
        c = box_intersection(a, b) / box_union(a, b);

    return c;
}

inline static int pad_up_16(int a)
{
    return ceil((float)a / 16) * 16;
}

// The data relocation is determined by the output_mem_addr2
int post_raw_info_process(int model_id, struct kdp_image_s *image_p)
{
    int grid_w, grid_h, grid_c, mem_offset;

    uint32_t *result_p = (uint32_t *)(POSTPROC_RESULT_MEM_ADDR(image_p));

    int idx_o = 0;
    result_p[idx_o++] = POSTPROC_OUTPUT_NUM(image_p);

    int idx = 0;
    int offset = sizeof(struct out_node_s);
    // save raw info
    struct out_node_s *out_p;
    for (idx = 0; idx < POSTPROC_OUTPUT_NUM(image_p); idx++)
    {
        out_p = (struct out_node_s *)((uint32_t)POSTPROC_OUT_NODE(image_p) + idx * offset);

        mem_offset = OUT_NODE_ADDR(out_p) - MODEL_OUTPUT_MEM_ADDR(image_p);
        grid_w = OUT_NODE_COL(out_p);
        grid_h = OUT_NODE_ROW(out_p);
        grid_c = OUT_NODE_CH(out_p);
        result_p[idx_o++] = POSTPROC_OUTPUT_MEM_ADDR(image_p) + mem_offset;
        result_p[idx_o++] = grid_h;
        result_p[idx_o++] = grid_c;
        result_p[idx_o++] = grid_w;
        result_p[idx_o++] = OUT_NODE_RADIX(out_p);
        result_p[idx_o++] = OUT_NODE_SCALE(out_p);
    }

    int dst_base = POSTPROC_OUTPUT_MEM_ADDR2(image_p);

    if (dst_base && ((image_p->raw_img_p->format & IMAGE_FORMAT_PARALLEL_PROC) == 0))
    {
        int data_size, len, grid_w_bytes_aligned, total, src_p, dst_p;
        data_size = (POSTPROC_OUTPUT_FORMAT(image_p) & BIT(0)) + 1; /* 1 or 2 in bytes */
        u32 block = POSTPROC_OUTPUT_NUM(image_p);
        idx_o = 1;

        for (idx = 0; idx < block; idx++)
        {
            out_p = (struct out_node_s *)((uint32_t)POSTPROC_OUT_NODE(image_p) + idx * offset);
            src_p = OUT_NODE_ADDR(out_p);
            mem_offset = dst_base - MODEL_OUTPUT_MEM_ADDR(image_p);
            dst_p = OUT_NODE_ADDR(out_p) + mem_offset;

            grid_w = OUT_NODE_COL(out_p);
            grid_h = OUT_NODE_ROW(out_p);
            grid_c = OUT_NODE_CH(out_p);

            len = grid_w * data_size;
            grid_w_bytes_aligned = pad_up_16(len);

            total = grid_h * grid_c * grid_w_bytes_aligned;
            kdrv_gdma_memcpy((uint32_t)dst_p, (uint32_t)src_p, total);
            result_p[idx_o] = dst_p;
            idx_o += 6;
        }
    }
    return 0;
}

static int8_t float_fo_fixp(float value, int radix)
{
    return (int8_t)(value * (1 << radix));
}

int user_post_yolo(int model_id, struct kdp_image_s *image_p)
{
    //print_kdp_image_s(image_p);

    struct yolo_v3_post_globals_s *gp = get_yolov3_gp();
    int i, j, k, div, ch, row, col, good_box_count, class_good_box_count, good_result_count, len;
    float box_x, box_y, box_w, box_h, box_confidence, max_score, *p_prob_thresh_yolov3;
    int8_t *src_p, *x_p, *y_p, *width_p, *height_p, *score_p, *class_p;
    struct bounding_box_s *bbox;
    struct yolo_result_s *result;
    struct bounding_box_s *result_box_p, *result_tmp_p, *r_tmp_p;

    int grid_w, grid_h, class_num, grid_w_bytes_aligned;
    uint32_t src_img_mode;

    float maxlen = RAW_INPUT_COL(image_p) > RAW_INPUT_ROW(image_p) ? (float)RAW_INPUT_COL(image_p) : (float)RAW_INPUT_ROW(image_p);

    p_prob_thresh_yolov3 = &((float *)POSTPROC_PARAMS_P(image_p))[YOLO_PROB_THRESH_POS];

    if (*p_prob_thresh_yolov3 > 0)
        prob_thresh_yolov3 = *p_prob_thresh_yolov3;

    src_img_mode = RAW_FORMAT(image_p);

    result = (struct yolo_result_s *)(POSTPROC_RESULT_MEM_ADDR(image_p));
    result_box_p = result->boxes;

    result_tmp_p = gp->result_tmp_s;

    class_num = POSTPROC_OUT_NODE_CH(image_p) / YOLO_V3_CELL_BOX_NUM - YOLO_BOX_FIX_CH;

    result->class_count = class_num;
    bbox = gp->bboxes_v3;
    good_box_count = 0;

    int node_idx;
    int offset = sizeof(struct out_node_s);
    struct out_node_s *out_p;

    for (node_idx = 0; node_idx < POSTPROC_OUTPUT_NUM(image_p); node_idx++)
    {
        uint32_t all_data;

        out_p = (struct out_node_s *)((uint32_t)POSTPROC_OUT_NODE(image_p) + node_idx * offset);
        src_p = (int8_t *)OUT_NODE_ADDR_PARALLEL(out_p, image_p);

        grid_w = OUT_NODE_COL(out_p);
        grid_h = OUT_NODE_ROW(out_p);
        len = grid_w;
        grid_w_bytes_aligned = round_up(len);
        len = grid_w_bytes_aligned;
        all_data = grid_w_bytes_aligned * (class_num + YOLO_BOX_FIX_CH);

        div = 1 << OUT_NODE_RADIX(out_p);

        int fxp_radix = OUT_NODE_RADIX(out_p);
        float fxp_scale = *(float *)&OUT_NODE_SCALE(out_p);
        float fScale = 1 / (div * fxp_scale);

        float temp_t1 = 1 / prob_thresh_yolov3 - 1;
        float temp_t2 = (float)(-log(temp_t1));
        float temp_t3 = fxp_scale * temp_t2;
        int8_t fxp_score_thresh = float_fo_fixp(temp_t3, fxp_radix);

        for (row = 0; row < grid_h; row++)
        {
            for (ch = 0; ch < YOLO_V3_CELL_BOX_NUM; ch++)
            {
                x_p = src_p;
                y_p = x_p + len;
                width_p = y_p + len;
                height_p = width_p + len;
                score_p = height_p + len;
                class_p = score_p + len;

                for (col = 0; col < grid_w; col++, x_p++, y_p++, width_p++, height_p++, score_p++, class_p++)
                {
                    if (*score_p <= fxp_score_thresh)
                        continue;

                    box_x = (float)*x_p;
                    box_y = (float)*y_p;
                    box_w = (float)*width_p;
                    box_h = (float)*height_p;

                    box_confidence = (float)*score_p;
                    box_confidence = sigmoid(do_div_scale_optim(box_confidence, fScale));

                    /* Get scores of all class */
                    for (i = 0; i < class_num; i++)
                        gp->box_class_probs[i] = (float)*(class_p + i * len);

                    /* Find all classes with score higher than thresh */
                    int done_box = 0;

                    for (i = 0; i < class_num; i++)
                    {
                        max_score = sigmoid(do_div_scale_optim(gp->box_class_probs[i], fScale)) * box_confidence;

                        if (max_score >= prob_thresh_yolov3)
                        {
                            if (!done_box)
                            {
                                done_box = 1;
                                box_x = do_div_scale_optim(box_x, fScale);
                                box_y = do_div_scale_optim(box_y, fScale);
                                box_w = do_div_scale_optim(box_w, fScale);
                                box_h = do_div_scale_optim(box_h, fScale);

                                if (model_id == KNERON_YOLOV5S_PersonBicycleCarMotorcycleBusTruckCatDog8_480_256_3)
                                {
                                    box_x = sigmoid(box_x);
                                    box_y = sigmoid(box_y);
                                    box_w = sigmoid(box_w);
                                    box_h = sigmoid(box_h);

                                    box_x = (box_x * 2 - 0.5f + col) / grid_w;
                                    box_y = (box_y * 2 - 0.5f + row) / grid_h;
                                    box_w *= 2;
                                    box_h *= 2;
                                    box_w = box_w * box_w * anchers_v5[node_idx][ch][0] / DIM_INPUT_COL(image_p);
                                    box_h = box_h * box_h * anchers_v5[node_idx][ch][1] / DIM_INPUT_ROW(image_p);

                                    bbox->x1 = (box_x - (box_w / 2)) * RAW_INPUT_COL(image_p);
                                    bbox->y1 = (box_y - (box_h / 2)) * RAW_INPUT_ROW(image_p);
                                    bbox->x2 = (box_x + (box_w / 2)) * RAW_INPUT_COL(image_p);
                                    bbox->y2 = (box_y + (box_h / 2)) * RAW_INPUT_ROW(image_p);
                                }
                                else // tiny yolo modles
                                {
                                    box_x = (sigmoid(box_x) + col) / grid_w;
                                    box_y = (sigmoid(box_y) + row) / grid_h;
                                    box_w = expf(box_w) * anchers_v3[node_idx][ch][0] / DIM_INPUT_COL(image_p);
                                    box_h = expf(box_h) * anchers_v3[node_idx][ch][1] / DIM_INPUT_ROW(image_p);

                                    if (src_img_mode & (uint32_t)IMAGE_FORMAT_CHANGE_ASPECT_RATIO)
                                    {
                                        bbox->x1 = (box_x - (box_w / 2)) * RAW_INPUT_COL(image_p);
                                        bbox->y1 = (box_y - (box_h / 2)) * RAW_INPUT_ROW(image_p);
                                        bbox->x2 = (box_x + (box_w / 2)) * RAW_INPUT_COL(image_p);
                                        bbox->y2 = (box_y + (box_h / 2)) * RAW_INPUT_ROW(image_p);
                                    }
                                    else
                                    {
                                        bbox->x1 = (box_x - (box_w / 2)) * maxlen;
                                        bbox->y1 = (box_y - (box_h / 2)) * maxlen;
                                        bbox->x2 = (box_x + (box_w / 2)) * maxlen;
                                        bbox->y2 = (box_y + (box_h / 2)) * maxlen;
                                    }
                                }
                            }
                            else
                            {
                                memcpy(bbox, bbox - 1, sizeof(struct bounding_box_s));
                            }

                            bbox->score = max_score;
                            bbox->class_num = i;

                            bbox++;
                            good_box_count++;

                            if (good_box_count >= YOLO_V3_O2_GRID_MAX * YOLO_V3_CELL_BOX_NUM)
                            {
                                kerr_msg("Allocate more memory for maximum good detection\n");
                                goto iou_box;
                            }
                        }
                    }
                }
                src_p += all_data;
            }
        }
    }

iou_box:

    kdbg_msg("  TY3: good_box_count %d\n", good_box_count);
    good_result_count = 0;

    for (i = 0; i < class_num; i++)
    {
        bbox = gp->bboxes_v3;
        class_good_box_count = 0;
        r_tmp_p = result_tmp_p;

        for (j = 0; j < good_box_count; j++)
        {
            if (bbox->class_num == i)
            {
                memcpy(r_tmp_p, bbox, sizeof(struct bounding_box_s));
                r_tmp_p++;
                class_good_box_count++;
            }
            bbox++;
        }

        if (class_good_box_count == 1)
        {
            memcpy(&result_box_p[good_result_count], &result_tmp_p[0], sizeof(struct bounding_box_s));
            good_result_count++;
        }
        else if (class_good_box_count >= 2)
        {
            qsort(result_tmp_p, class_good_box_count, sizeof(struct bounding_box_s), box_score_comparator);
            for (j = 0; j < class_good_box_count; j++)
            {
                if (result_tmp_p[j].score == 0)
                    continue;
                for (k = j + 1; k < class_good_box_count; k++)
                {
                    if (box_iou(&result_tmp_p[j], &result_tmp_p[k]) > nms_thresh_yolov3_ud)
                    {
                        result_tmp_p[k].score = 0;
                    }
                }
            }

            int good_count = 0;
            for (j = 0; j < class_good_box_count; j++)
            {
                if (result_tmp_p[j].score > 0)
                {
                    memcpy(&result_box_p[good_result_count], &result_tmp_p[j], sizeof(struct bounding_box_s));
                    good_result_count++;
                    good_count++;
                }
                if (YOLO_MAX_DETECTION_PER_CLASS == good_count || YOLO_GOOD_BOX_MAX == good_result_count)
                {
                    break;
                }
            }
        }
    }

    for (i = 0; i < good_result_count; i++)
    {
        result->boxes[i].x1 = (int)(result->boxes[i].x1 + (float)0.5) < 0 ? 0 : (int)(result->boxes[i].x1 + (float)0.5);
        result->boxes[i].y1 = (int)(result->boxes[i].y1 + (float)0.5) < 0 ? 0 : (int)(result->boxes[i].y1 + (float)0.5);
        result->boxes[i].x2 = (int)(result->boxes[i].x2 + (float)0.5) > RAW_INPUT_COL(image_p) ? RAW_INPUT_COL(image_p) : (int)(result->boxes[i].x2 + (float)0.5);
        result->boxes[i].y2 = (int)(result->boxes[i].y2 + (float)0.5) > RAW_INPUT_ROW(image_p) ? RAW_INPUT_ROW(image_p) : (int)(result->boxes[i].y2 + (float)0.5);
    }

    kdbg_msg("good_result_count: %d\n", good_result_count);
    result->box_count = good_result_count;
    for (unsigned int i = 0; i < good_result_count; i++)
    {
        kdbg_msg("post_yolo3 %f %f %f %f %f %d\n", result->boxes[i].x1, result->boxes[i].y1, result->boxes[i].x2, result->boxes[i].y2, result->boxes[i].score, result->boxes[i].class_num);
    }
    len = good_result_count * sizeof(struct bounding_box_s);

    return len;
}
