/*
 * Kneron Application general functions
 *
 * Copyright (C) 2021 Kneron, Inc. All rights reserved.
 *
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "kmdw_console.h"

#include "kmdw_inference_app.h"
#include "kdp2_inf_app_yolo.h"

void kdp2_app_yolo_result_callback(int status, void *inf_result_buf, int inf_result_buf_size, void *ncpu_result_buf)
{
    // when ncpu has done post-process then it comes to here with inference result buffer filled
    // 'inf_result_buf' is used to carry inference result back to host SW = header + inferernce result (from ncpu/npu)
    // 'ncpu_result_buf' is post-processing result buffer done by ncpu
    // normally 'inf_result_buf' contains 'ncpu_result_buf', and user should send 'inf_result_buf' to host SW

    kdp2_ipc_app_yolo_result_t *app_yolo_result = (kdp2_ipc_app_yolo_result_t *)inf_result_buf;
    kp_app_yolo_result_t *yolo_data = (kp_app_yolo_result_t *)ncpu_result_buf;

    if (status != KP_SUCCESS)
    {
        app_yolo_result->header_stamp.status_code = status;
        app_yolo_result->header_stamp.total_size = sizeof(kdp2_ipc_app_yolo_result_t) - sizeof(kp_app_yolo_result_t);
        // send error status result back to host
        kmdw_inference_app_result_enqueue((void *)inf_result_buf, inf_result_buf_size, false);
        return;
    }

    int box_cnt = yolo_data->box_count;
    if (box_cnt > YOLO_GOOD_BOX_MAX)
    {
        kmdw_printf("[app_yolo]: error ! too many bounding boxes = %d!!!\n", box_cnt);
        box_cnt = YOLO_GOOD_BOX_MAX;
    }

    app_yolo_result->header_stamp.status_code = KP_SUCCESS;
    app_yolo_result->header_stamp.total_size = sizeof(kdp2_ipc_app_yolo_result_t) - sizeof(kp_app_yolo_result_t) +
                                               sizeof(yolo_data->class_count) + sizeof(yolo_data->box_count) + box_cnt * sizeof(kp_bounding_box_t);

    // send output result buffer back to host SW
    kmdw_inference_app_result_enqueue((void *)inf_result_buf, inf_result_buf_size, false);
}

void kdp2_app_yolo_inference(uint32_t job_id, void *inf_input_buf)
{
    // 'inf_input_buf' and 'inf_result_buf' are provided by kdp2 middleware
    // the content of 'inf_input_buf' is transmitted from host SW = header + image
    // 'inf_result_buf' is used to carry inference result back to host SW = header + inferernce result (from ncpu/npu)

    // now get an available free result buffer
    // normally the begin part of result buffer should contain app-defined result header
    // and the rest is for ncpu/npu inference output data
    int result_buf_size;
    void *inf_result_buf = kmdw_inference_app_result_get_free_buffer(&result_buf_size);

    kdp2_ipc_app_yolo_inf_header_t *app_yolo_header = (kdp2_ipc_app_yolo_inf_header_t *)inf_input_buf;
    kdp2_ipc_app_yolo_result_t *app_yolo_result = (kdp2_ipc_app_yolo_result_t *)inf_result_buf;

    // config image preprocessing and model settings
    kmdw_inference_app_config_t inf_config;
    memset(&inf_config, 0, sizeof(kmdw_inference_app_config_t)); // for safety let default 'bool' to 'false'

    // image buffer address should be just after the header
    inf_config.image_buf = (void *)((uint32_t)app_yolo_header + sizeof(kdp2_ipc_app_yolo_inf_header_t));
    inf_config.image_width = app_yolo_header->width;
    inf_config.image_height = app_yolo_header->height;
    inf_config.image_channel = app_yolo_header->channel;
    inf_config.image_format = app_yolo_header->image_format;
    inf_config.image_norm = app_yolo_header->model_normalize;
    inf_config.image_resize = KP_RESIZE_ENABLE;
    inf_config.image_padding = KP_PADDING_CORNER;
    inf_config.model_id = app_yolo_header->model_id;
    // inf_config.enable_crop = false;                                     // no support crop
    // inf_config.enable_raw_output = false;                               // we have post-process for this app
    inf_config.enable_parallel = true;                                  // only works for single model and post-process in ncpu
    inf_config.inf_result_buf = inf_result_buf;                         // for callback
    inf_config.inf_result_buf_size = result_buf_size;                   //
    inf_config.ncpu_result_buf = (void *)&(app_yolo_result->yolo_data); // give result buffer for ncpu/npu, callback will carry it
    inf_config.result_callback = kdp2_app_yolo_result_callback;
    inf_config.pad_value = NULL;
    inf_config.user_define_data = NULL;

    // pre-set something for result output
    // header_stamp is a must to correctly transfer result data back to host SW
    app_yolo_result->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
    app_yolo_result->header_stamp.total_size = 0;
    app_yolo_result->header_stamp.job_id = job_id;
    app_yolo_result->inf_number = app_yolo_header->inf_number; // sync the inference number

    // run preprocessing and inference, trigger ncpu/npu to do the work
    // if enable_parallel=true (works only for single model), result callback is needed
    // however if inference error then no callback will be invoked
    int ret = kmdw_inference_app_execute(&inf_config);
    if (ret != KP_SUCCESS)
    {
        // some sort of inference error
        app_yolo_result->header_stamp.status_code = ret;
        app_yolo_result->header_stamp.total_size = sizeof(kdp2_ipc_app_yolo_result_t) - sizeof(kp_app_yolo_result_t);

        // send error status result back to host
        kmdw_inference_app_result_enqueue((void *)inf_result_buf, result_buf_size, false);
    }
}
